
!!$ OFF2ANG2d - Conversion from offset to angle and back
!!$
!!$=head1 SYNOPSIS
!!$
!!$OFF2ANG par=parfile.P < input.H > output.H 
!!$
!!$=head1 INPUT PARAMETERS
!!$
!!$=over 4
!!$
!!$=item inverse=.false. (logical)  
!!$
!!$      Whether to do forward (offset to  angle) 
!!$                 or inverse ( angle to offset) transformation
!!$
!!$=item maxmem=100. (real) 
!!$
!!$      Maximum in-core memory (in Mbytes)
!!$
!!$=item eps=1.0 (real) 
!!$
!!$      Epsilon parameter used by the radial trace transform
!!$
!!$=item push=.false. (logical)
!!$
!!$      push=.false. means pull transformation
!!$
!!$=item verb=.false. (logical) 
!!$
!!$      Verbose execution
!!$
!!$=item amax=float  (maximum angle)
!!$
!!$      Maximum angle
!!$
!!$=item nzpad=n1 (int) 
!!$
!!$      Pading on the depth axis
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Conversion from angle to offset and back.
!!$ Uses radial trace transforms in the Fourier domain.
!!$ Theory in Sava & Fomel(2000) - SEP103
!!$
!!$ real input/output
!!$
!!$  forward transformation
!!$  input  is z-hx-mx(-ro)
!!$  output is z-ax-mx(-ro)
!!$  
!!$  inverse transformation
!!$  input  is z-ax-mx(-ro)
!!$  output is z-hx-mx(-ro)
!!$
!!$=head1 KEYWORDS
!!$
!!$  angle-domain, wave-equation
!!$
!!$=head1 CATEGORY
!!$
!!$B<seis/image>
!!$
!!$=cut
!!$
!!$ Paul Sava (paul@sep.stanford.edu)
!!$ 01/10/01
!!$
!!$
!>
!%%
program Off2ang2d

  use sep
  use utilities_mod
  use omppar_mod
  use adcig_mod
  use odcig_mod

  integer, parameter:: dbl=selected_int_kind(15)

  type(paxis) :: az,amx,ar
  type(paxis) :: ahx
  type(paxis) :: aax

  integer, external :: omp_get_num_threads
  integer, external :: omp_get_thread_num

  real        :: amax,hmax

  real        :: eps,pi
  logical     :: verb, push
  integer     :: nzpad,nzold
  integer     :: stat,i1,i2,i3

  !! z-hx-hy-ngathers
  real,    dimension(:,:,:), pointer :: rdata
  complex, dimension(:,:,:), pointer :: cdata

  logical :: inverse
  integer :: esize
  real    :: maxmem    !! max memory available  (Mb)
  real    :: gathermem !! memory for one gather (Mb)
  integer :: nth,      ith
  integer :: nblocks,  iblocks
  integer(kind=dbl) :: ngathers, igathers, allgathers

  !$OMP PARALLEL
  nth = omp_get_num_threads()
  !$OMP END PARALLEL

  pi = acos(-1.)
!  call sep_init("/opt/SEP/seplib/seplib_prog/seis/image/OFF2ANG.f90s")

  call sep_init()
  !----------------------------------------------------------------
  call from_history("esize",esize,0)
  if(esize /=4) call seperr("need real input data (esize=4)")
  !----------------------------------------------------------------

  call pinaxis(az ,1)
  call pinaxis(amx,3)
  call pinaxis(ar ,4)

  call from_param("inverse",inverse,.false.)
  call from_param("maxmem",maxmem,   100.)    !! Mb
  call from_param(   "eps",   eps,    1.0)
  call from_param(  "push",  push,.false.)
  call from_param(  "verb",  verb,.false.)

  call from_param( "nzpad", nzpad,   az%n)
  nzold = az%n
  az%n = nzpad

  if(inverse) then
     !----------------------------------------------------------------
     call pinaxis(aax,2)
     if(aax%n==1) call seperr("need prestack data")
     !----------------------------------------------------------------
     call from_param(  "hmax",  hmax,    1.) ! change this
     !----------------------------------------------------------------

     ahx%n = aax%n   
     ahx%d = 1.
     if(ahx%n /=1) then
        if(aax%o ==0.) then
           ahx%o = 0.
           ahx%d =   hmax/ahx%n
        else
           ahx%o =  -hmax
           ahx%d = 2*hmax/ahx%n
        end if
        call pouaxis(ahx,2)
        call to_history ("label2","hx")

        if(verb) then
           write(0,*) "requested max hx =",hmax
        end if
     end if

     !----------------------------------------------------------------
  else
     !----------------------------------------------------------------
     call pinaxis(ahx,2); if(ahx%d==0.) ahx%d=1.
     if(ahx%n==1) call seperr("need prestack data")
     !----------------------------------------------------------------
     call from_param(  "amax",  amax,    60.)
     !----------------------------------------------------------------

     aax%n = ahx%n
     if(ahx%n /=1) then
        if(ahx%o ==0.) then
           aax%o = 0.
           aax%d =   amax/aax%n
        else
           aax%o =  -amax
           aax%d = 2*amax/aax%n
        end if
        call pouaxis(aax,2)
        call to_history ("label2","ax")

        if(verb) then
           write(0,*) "requested max ax =",amax
           write(0,*) "believe  only ax <",180*atan(az%d/ahx%d)/pi
        end if
     end if

     !----------------------------------------------------------------
  end if

  call sep_close()

  !----------------------------------------------------------------
  !! find out how many threads the environment variable actually allows
  !$OMP PARALLEL
  nth = omp_get_num_threads()
  !$OMP END PARALLEL

  !! total # of gathers
  allgathers = ar%n * amx%n

  !! memory needed for one gather
  if(inverse) then
     gathermem = (1/1024./1024.) * az%n * aax%n * 12 
     !! z*ax*(4+8) (Mb)
  else
     gathermem = (1/1024./1024.) * az%n * ahx%n * 12 
     !! z*hx*(4+8) (Mb)
  end if

  !! find the # of gathers that fit in the allocated memory
  ngathers = maxmem/gathermem

  !! find, if possible, the max value of ngathers 
  !!       which divides allgathers
  !!      and is divided by nth
  call omppar(allgathers,ngathers,nth)

  !! set and freeze the # of threads
  call omp_set_dynamic(.false.)
  call omp_set_num_threads(nth)

  !! # of data blocks
  nblocks = allgathers/ngathers
  !----------------------------------------------------------------

  if(verb) then
     write(0,*) "   inverse=",inverse
     write(0,*) "      push=",push
     write(0,*) "    maxmem=",maxmem,'Mb'
     write(0,*) "    effmem=",ngathers*gathermem,'Mb'
     write(0,*) "allgathers=",allgathers
     write(0,*) "  ngathers=",ngathers
     write(0,*) "       nth=",nth
     write(0,*) "   nblocks=",nblocks
  end if
  !----------------------------------------------------------------  

  if(inverse) then
     allocate(rdata(nzold,aax%n,ngathers), stat=stat)
     if(stat /=0) call seperr("cant allocate rdata")
     allocate(cdata(nzpad,aax%n,ngathers), stat=stat)
     if(stat /=0) call seperr("cant allocate cdata")
  else
     allocate(rdata(nzold,ahx%n,ngathers), stat=stat)
     if(stat /=0) call seperr("cant allocate rdata")
     allocate(cdata(nzpad,ahx%n,ngathers), stat=stat)
     if(stat /=0) call seperr("cant allocate cdata")
  end if
  rdata =     0.
  cdata = (0.,0.)

  !----------------------------------------------------------------

  !! reserve enough disk space for the entire output
  if(verb) write(0,*) "reserving disk space"
  do iblocks=1,nblocks
     call sep_write(rdata)
  end do
  call sseek('out',0,0)
  if(verb) write(0,*) "OK"

  call sep_begin_prog()


  !----------------------------------------------------------------
  if(inverse) then
     if(verb) write(0,*) "compute offsets"
     call odcig_fft_init(nth,(/az%n,ahx%n/))
     call odcig_shift_init(az,ahx)

     !----------------------------------------------------------------  
     do iblocks=1,nblocks
        if(verb) write(0,*) iblocks,'/',nblocks

        call sep_read(rdata)
        do i3=1,size(cdata,3)
        do i2=1,size(cdata,2)
        do i1=1,nzold
         cdata(i1,i2,i3) = cmplx(rdata(i1,i2,i3),0.)
        end do;end do;end do;


        !----------------------------------------------------------------
        !$OMP  PARALLEL
        !$OMP  DO &
        !$OMP& SCHEDULE(DYNAMIC) &
        !$OMP& PRIVATE(ith,igathers)
        do igathers=1,ngathers

           ith = omp_get_thread_num()+1

           call odcig_F_fft(ith,igathers,cdata)
           call odcig_F_shift(  igathers,cdata)
        end do
        !$OMP END DO
        !$OMP END PARALLEL
        !----------------------------------------------------------------
        if(aax%n >1) then
           call odcig_off_init(nth,az,ahx,aax,eps,push)

           !$OMP  PARALLEL
           !$OMP  DO &
           !$OMP& SCHEDULE(DYNAMIC) &
           !$OMP& PRIVATE(ith,igathers)
           do igathers=1,ngathers
              ith = omp_get_thread_num()+1
              call odcig_off_x(ith,igathers,cdata,az,ahx,aax)
           end do
           !$OMP END DO
           !$OMP END PARALLEL

           call odcig_off_close()
        end if

        !----------------------------------------------------------------
        !$OMP  PARALLEL
        !$OMP  DO &
        !$OMP& SCHEDULE(DYNAMIC) &
        !$OMP& PRIVATE(ith,igathers)
        do igathers=1,ngathers
           ith = omp_get_thread_num()+1

           call odcig_I_shift(  igathers,cdata)
           call odcig_I_fft(ith,igathers,cdata)
        end do
        !$OMP END DO
        !$OMP END PARALLEL
        !----------------------------------------------------------------

write(0,*) "DATA",sum(cdata)
        rdata(1:nzold,:,:) = real (cdata(1:nzold,:,:))
        call sep_write(rdata)

     end do !! iblocks
     !----------------------------------------------------------------

     call odcig_shift_close()
     call odcig_fft_close()
  else
     if(verb) write(0,*) "compute angles"
     call adcig_fft_init(nth,(/az%n,aax%n/))
     call adcig_shift_init(az,ahx)

     !----------------------------------------------------------------  
     do iblocks=1,nblocks
        if(verb) write(0,*) iblocks,'/',nblocks

        call sep_read(rdata)
        do i3=1,size(cdata,3)
        do i2=1,size(cdata,2)
        do i1=1,nzold
         cdata(i1,i2,i3) = cmplx(rdata(i1,i2,i3),0.)
        end do;end do;end do;

        !----------------------------------------------------------------
        !$OMP  PARALLEL
        !$OMP  DO &
        !$OMP& SCHEDULE(DYNAMIC) &
        !$OMP& PRIVATE(ith,igathers)
        do igathers=1,ngathers
           ith = omp_get_thread_num()+1

           call adcig_F_fft(ith,igathers,cdata)
           call adcig_F_shift(  igathers,cdata)
        end do
        !$OMP END DO
        !$OMP END PARALLEL
        !----------------------------------------------------------------
        if(ahx%n >1) then
           call adcig_ang_init(nth,az,ahx,aax,eps,push)

           !$OMP  PARALLEL
           !$OMP  DO &
           !$OMP& SCHEDULE(DYNAMIC) &
           !$OMP& PRIVATE(ith,igathers)
           do igathers=1,ngathers
              ith = omp_get_thread_num()+1
              call adcig_ang_x(ith,igathers,cdata,az,ahx,aax)
           end do
           !$OMP END DO
           !$OMP END PARALLEL

           call adcig_ang_close()
        end if
  
        !----------------------------------------------------------------
        !$OMP  PARALLEL
        !$OMP  DO &
        !$OMP& SCHEDULE(DYNAMIC) &
        !$OMP& PRIVATE(ith,igathers)
        do igathers=1,ngathers
           ith = omp_get_thread_num()+1

           call adcig_I_shift(  igathers,cdata)
           call adcig_I_fft(ith,igathers,cdata)
        end do
        !$OMP END DO
        !$OMP END PARALLEL
        !----------------------------------------------------------------


        do i3=1,size(rdata,3)
        do i2=1,size(rdata,2)
        do i1=1,nzold
        rdata(i1,i2,i3) = real (cdata(i1,i2,i3))
       end do;end do;end do;
        call sep_write(rdata)

     end do !! iblocks
     !----------------------------------------------------------------  

     call adcig_shift_close()
     call adcig_fft_close()
  end if

  deallocate(rdata, stat=stat)
  if(stat /=0) call seperr("cant deallocate rdata")
  deallocate(cdata, stat=stat)
  if(stat /=0) call seperr("cant deallocate cdata")
  call sep_end_prog()

end program 


