program TiltCorScan
!! Do a scanning over a range of alpha, beta, and gamma that minimize the tilt correction creteria
!! Input: nalpha, nbeta, ngamma
!! Input: oalpha, obeta, ogamma
!! Input: dalpha, dbeta, dgamma
!! Input: CRG (Sx, Sy, time, components)
!! Output Energy file (nalpha, nbeta, ngamma)
!! minimize   sum_t || abs(P) - Calib*abs(V) ||^2
!! where Calib = calibration factor ~ avg(P)/avg(V)


   use sep
   use type_mod
   use rotation_mod
   
   implicit none
   
   type(rotcoeff)        ::coe
   type(comvec)         :: vin, vout
   
   integer               :: nt, nsx, it, isx, n1, n2
   integer               :: ialpha, ibeta, igam
   integer               :: nalpha, nbeta, ngam
   real                  :: dalpha, dbeta, dgam
   real                  :: oalpha, obeta, ogam
   real                  :: ot, dt, osx, dsx
   real                  :: alpha, beta, gam, pi, cost, Result(3), Calib
   character(len=128)    :: Pfile_, xfile_, yfile_, zfile_, Rotfile_
   character(len=128)    :: Objfile_
   real, dimension(:,:), allocatable  :: Parray, Xarray, Yarray, Zarray
   real, dimension(:,:), allocatable  :: XarryOut, YarryOut, VarryOut   
   real, dimension(:,:,:), allocatable  ::   Eout
   
    !! For OMP
  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer, external  :: srite_window, sreed_window, getch
  integer            :: inode1, inode, nodes
   
   
   call sep_init()
   
  !!======================================================
  !! Parameter reading
  
  call from_param("nalpha", nalpha,10);   call from_param("dalpha",dalpha,1.0); call from_param("oalpha",oalpha,-5.0);
  call from_param("nbeta", nbeta,10);   call from_param("dbeta",dbeta,1.0); call from_param("obeta",obeta,-5.0);
  call from_param("ngamma", ngam,10);   call from_param("dgamma",dgam,1.0); call from_param("ogamma",ogam,-5.0);
  call from_param("nodes",nodes, 4);
  call from_param("Calib",Calib);
      
  call from_param("Pfile", Pfile_);  
  call from_param("Xfile", Xfile_);
  call from_param("Yfile", Yfile_);  
  call from_param("Zfile", Zfile_); 
  call from_param("Rotfile", Rotfile_,"Rotfile.H")   
  
  ! Parallel Programming

  if (nodes .gt. omp_get_num_procs()) nodes =omp_get_num_procs()-3;
    call omp_set_num_threads(nodes);
    inode1=0
  
  write(0,*) '***********************'
  write(0,*) 'alpha: n, d, o', nalpha, dalpha, oalpha
  write(0,*) 'beta: n, d, o', nbeta, dbeta, obeta
  write(0,*) 'gama: n, d, o', ngam, dgam, ogam
  write(0,*) 'Calib', Calib
  write(0,*) '***********************'
  
  call from_aux(Pfile_,"n1", nt); call from_aux(Pfile_,"n2", nsx); 
  call from_aux(Pfile_,"d1", dt); call from_aux(Pfile_,"d2", dsx); 
  call from_aux(Pfile_,"o1", ot); call from_aux(Pfile_,"o2", osx); 
  call from_aux(Xfile_,"n1", n1); call from_aux(Pfile_,"n2", n2);  
  if ((n1 .ne. nt) .or. (n2 .ne. nsx)) call erexit('dimension does not agree between Xfile and Pfile')
  call from_aux(Yfile_,"n1", n1); call from_aux(Yfile_,"n2", n2);  
  if (n1 .ne. nt) call erexit('dimension  does not agree between Yfile and Pfile')
  call from_aux(Zfile_,"n1", n1); call from_aux(Zfile_,"n2", n2);  
  if (n1 .ne. nt) call erexit('dimension  does not agree between Zfile and Pfile')    
   
  ! Read in common receiver gathers
  allocate(Parray(nt,nsx), xarray(nt,nsx), yarray(nt,nsx), zarray(nt,nsx))
  allocate(XarryOut(nt,nsx), YarryOut(nt,nsx), VarryOut(nt,nsx))  
  call sep_read(Parray,Pfile_);
  call sep_read(Xarray,Xfile_);
  call sep_read(Yarray,Yfile_);
  call sep_read(Zarray,Zfile_);       
   
  
  pi = 3.14156 
  allocate(Eout(nalpha,nbeta,ngam))
  Eout = 0.! Initialize  

if (1 == 1) then
  
  ! Scan over all the alpha, beta, gama values to find the minimum
  !$OMP parallel do ordered private(inode, igam, ibeta, ialpha, gam, beta, alpha, coe, cost) schedule(dynamic,1)
  do igam=1,ngam
  
    inode=omp_get_thread_num(); 
    if(nodes==0.) inode=0; 
    if(inode>=inode1) inode1=inode;
  
    gam = ((igam-1)*dgam + ogam)*pi/180.
    do ibeta=1, nbeta
      write(0,*) 'igam=',igam,'ibeta =', ibeta
      beta = ((ibeta-1)*dbeta + obeta)*pi/180.
      do ialpha=1,nalpha
         alpha = ((ialpha-1)*dalpha + oalpha)*pi/180.
         
         call CoefCal(alpha,beta,gam,coe)
         call ObjectFun(nt,nsx,Parray, Xarray, Yarray, Zarray, Calib, coe, cost)

        !$OMP CRITICAL
         Eout(ialpha,ibeta,igam) = cost
        !$OMP END CRITICAL
        
        
              
      end do
    end do
  end do 
  !$OMP end parallel do  
   
end if
  


  !-------------Calculate the Rotated Components  
  Result = minloc(Eout)
  write(0,*) 'Min locatoin of E', Result
  alpha = (Result(1)-1.0)*dalpha + oalpha
  beta = (Result(2)-1.0)*dbeta + obeta
  gam = (Result(3)-1.0)*dgam + ogam
  write(0,*) 'In degree the minimum is', alpha, beta, gam
  
  call CoefCal(alpha,beta,gam,coe)
  do it=1,nt
     do isx=1,nsx
       vin%x = Xarray(it,isx)
       vin%y = Yarray(it,isx)
       vin%z = Zarray(it,isx) 
       call RotateXYV(vout,vin, coe)
       XarryOut(it,isx) = vout%x
       YarryOut(it,isx) = vout%y
       VarryOut(it,isx) = vout%z              
     end do
  end do   

  call sep_write(Parray, Rotfile_ , dim1=nt, dim2=nsx, esize=4)
  call sep_write(VarryOut, Rotfile_ , dim1=nt, dim2=nsx, esize=4)
  call sep_write(XarryOut, Rotfile_ , dim1=nt, dim2=nsx, esize=4)
  call sep_write(YarryOut, Rotfile_ , dim1=nt, dim2=nsx, esize=4)  
  
  call auxputch("n1",'i',nt, trim(Rotfile_)); call auxputch("d1",'f', dt, trim(Rotfile_));  call auxputch("o1",'f', ot, trim(Rotfile_));
   call auxputch("n2",'i', nsx, trim(Rotfile_)); call auxputch("d2",'f', dsx, trim(Rotfile_));  call auxputch("o2",'f', osx, trim(Rotfile_));
   call auxputch("n3",'i', 4, trim(Rotfile_)); call auxputch("d3",'f', 1., trim(Rotfile_));  call auxputch("o3",'f', 0., trim(Rotfile_));               
   call auxputch("label1", 's', 'nt', trim(Rotfile_));
   call auxputch("label2", 's', 'nsx', trim(Rotfile_));
   call auxputch("label3", 's', 'PVXY', trim(Rotfile_));
  
  !------------
  ! Output
  
  call sep_write(Eout, dim1=nalpha, dim2=nbeta, dim3=ngam, esize=4)
  
   call to_history("n1", nalpha);  call to_history("d1", dalpha);  call to_history("o1",oalpha);
   call to_history("n2", nbeta);  call to_history("d2", dbeta);  call to_history("o2",obeta);
   call to_history("n3", ngam);  call to_history("d3", dgam);  call to_history("o3",ogam);
   call to_history("label1",'alpha');      call to_history("label2",'beta'); 
      call to_history("label3",'gamma'); 
  
   
   



end program
